import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

HOST_PARAMETERS_PATH = r'Services\Tcpip\Parameters'


class HostDomainNamePlugin(Plugin):
    NAME = 'host_domain_name'
    DESCRIPTION = 'Get the computer host and domain names'
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug('Started Host and Domain Name Plugin...')

        for subkey_path in self.registry_hive.get_control_sets(HOST_PARAMETERS_PATH):
            subkey = self.registry_hive.get_key(subkey_path)

            hostname = subkey.get_value('Hostname', as_json=self.as_json)
            domain = subkey.get_value('Domain', as_json=self.as_json)

            # The default key value is 0x00000000 (REG_DWORD) when
            # the Windows machine is not in an AD domain.
            if not isinstance(domain, str):
                domain = None

            self.entries.append({
                'hostname': hostname,
                'domain': domain,
                'timestamp': convert_wintime(subkey.header.last_modified, as_json=self.as_json)
            })





import logging

from inflection import underscore

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

LAST_LOGON_KEY_PATH = r'\Microsoft\Windows\CurrentVersion\Authentication\LogonUI'


class LastLogonPlugin(Plugin):
    NAME = 'last_logon_plugin'
    DESCRIPTION = 'Get the last logged on username'
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        try:
            subkey = self.registry_hive.get_key(LAST_LOGON_KEY_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(f'Could not find {self.NAME} subkey at {LAST_LOGON_KEY_PATH}: {ex}')
            return None

        self.entries = {
            'last_write': convert_wintime(subkey.header.last_modified, as_json=self.as_json),
            **{underscore(x.name): x.value for x in subkey.iter_values(as_json=self.as_json)}
        }





import logging

from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import get_subkey_values_from_list

logger = logging.getLogger(__name__)


class TemplatePlugin(Plugin):
    NAME = 'template_plugin'
    DESCRIPTION = 'template_description'

    def can_run(self):
        # TODO: Choose the relevant condition - to determine if the plugin is relevant for the given hive
        return self.registry_hive.hive_type == NTUSER_HIVE_TYPE

    def run(self):
        # TODO: Return the relevant values
        raise NotImplementedError



import datetime
from regipy.exceptions import RegistryKeyNotFoundException

import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

INTERFACES_PATH = r'Services\Tcpip\Parameters\Interfaces'

class NetworkDataPlugin(Plugin):
    NAME = 'network_data'
    DESCRIPTION = 'Get network data from many interfaces'
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def get_network_info(self, subkey, interfaces=None):
        if interfaces is None:
            interfaces = []

        for interface in subkey.iter_subkeys():
            entries = {
                "interface_name": interface.name,
                "last_modified": convert_wintime(interface.header.last_modified, as_json=self.as_json),
                "dhcp_enabled": interface.get_value("EnableDHCP") == 1,  # Boolean value
            }

            if entries["dhcp_enabled"]:
                entries.update({
                    "dhcp_server": interface.get_value("DhcpServer"),
                    "dhcp_ip_address": interface.get_value("DhcpIPAddress"),
                    "dhcp_subnet_mask": interface.get_value("DhcpSubnetMask"),
                    "dhcp_default_gateway": interface.get_value("DhcpDefaultGateway"),
                    "dhcp_name_server": interface.get_value("DhcpNameServer"),
                    "dhcp_domain": interface.get_value("DhcpDomain"),
                })

                lease_obtained_time = interface.get_value("LeaseObtainedTime")
                if lease_obtained_time is not None:
                    lease_obtained_time_str = datetime.datetime.utcfromtimestamp(lease_obtained_time).strftime("%Y-%m-%d %H:%M:%S")
                    entries["dhcp_lease_obtained_time"] = lease_obtained_time_str

                lease_terminates_time = interface.get_value("LeaseTerminatesTime")
                if lease_terminates_time is not None:
                    lease_terminates_time_str = datetime.datetime.utcfromtimestamp(lease_terminates_time).strftime("%Y-%m-%d %H:%M:%S")
                    entries["dhcp_lease_terminates_time"] = lease_terminates_time_str

            else:
                entries.update({
                    "ip_address": interface.get_value("IPAddress"),
                    "subnet_mask": interface.get_value("SubnetMask"),
                    "default_gateway": interface.get_value("DefaultGateway"),
                    "name_server": interface.get_value("NameServer"),
                    "domain": interface.get_value("Domain"),
                })

            if interface.subkey_count:
                sub_interfaces = []
                self.get_network_info(self, interface, sub_interfaces)
                entries["sub_interface"] = sub_interfaces

            interfaces.append(entries)

        return interfaces
        
    def run(self):
        self.entries = {}

        for control_set_interfaces_path in self.registry_hive.get_control_sets(INTERFACES_PATH):
            try:
                subkey = self.registry_hive.get_key(control_set_interfaces_path)
            except RegistryKeyNotFoundException as ex:
                logger.error(ex)
                continue

            self.entries[control_set_interfaces_path] = {
                'timestamp': convert_wintime(subkey.header.last_modified, as_json=self.as_json)
            }
            interfaces = []
            interfaces = self.get_network_info(subkey, interfaces)
            self.entries[control_set_interfaces_path]['interfaces'] = interfaces



import binascii
import csv
import json
import logging
import os
import time
from typing import Generator, Iterator 


import attr
import click
import pytz 
from tabulate import tabulate

from regipy.plugins.plugin import PLUGINS
from regipy.recovery import apply_transaction_logs
from regipy.regdiff import compare_hives
from regipy.plugins.utils import run_relevant_plugins
from regipy.registry import NKRecord, RegistryHive
from regipy.exceptions import RegistryKeyNotFoundException
from regipy.utils import calculate_xor32_checksum, _setup_logging
from regipy.cli_utils import get_filtered_subkeys, _normalize_subkey_fields

logger = logging.getLogger(__name__)


@click.command()
@click.argument('hive_path', type=click.Path(exists=True, dir_okay=False, resolve_path=True), required=True)
@click.option('-v', '--verbose', is_flag=True, default=True, help='Verbosity')
def parse_header(hive_path, verbose):
    _setup_logging(verbose=verbose)
    registry_hive = RegistryHive(hive_path)

    click.secho(tabulate(registry_hive.header.items(), tablefmt='fancy_grid'))

    if registry_hive.header.primary_sequence_num != registry_hive.header.secondary_sequence_num:
        click.secho('Hive is not clean! You should apply transaction logs', fg='red')

    calculated_checksum = calculate_xor32_checksum(registry_hive._stream.read(508))
    if registry_hive.header.checksum != calculated_checksum:
        click.secho('Hive is not clean! Header checksum does not match', fg='red')


@click.command()
@click.argument('hive_path', type=click.Path(exists=True, dir_okay=False, resolve_path=True), required=True)
@click.option('-o', 'output_path', type=click.Path(exists=False, dir_okay=False, resolve_path=True), required=False)
@click.option('-p', '--registry-path', help='A registry path to start iterating from')
@click.option('-t', '--timeline', is_flag=True, default=False, help='Create a CSV timeline instead')
@click.option('-l', '--hive-type', type=click.STRING, required=False,
              help='Specify a hive type, if it could not be identified for some reason')
@click.option('-r', '--partial_hive_path', type=click.STRING, required=False,
              help='The path from which the partial hive actually starts, for example: -t ntuser -r "/Software" '
                   'would mean this is actually a HKCU hive, starting from HKCU/Software')
@click.option('-v', '--verbose', is_flag=True, default=False, help='Verbosity')
@click.option('-d', '--do-not-fetch-values', is_flag=True, default=False, help='Not fetching the values for each subkey '
                                                                              'makes the iteration way faster. Values count will still be returned')
@click.option('-s', '--start-date', type=click.STRING, required=False,
              help='If "-s" was specified, fetch only values for subkeys starting this timestamp in isoformat')
@click.option('-e', '--end-date', type=click.STRING, required=False,
              help='If "-e" was specified, fetch only values for subkeys until this timestamp in isoformat')
def registry_dump(hive_path, output_path, registry_path, timeline, hive_type, partial_hive_path, verbose, do_not_fetch_values, start_date, end_date):
    _setup_logging(verbose=verbose)
    registry_hive = RegistryHive(hive_path, hive_type=hive_type, partial_hive_path=partial_hive_path)

    start_time = time.monotonic()

    if registry_path:
        try:
            name_key_entry = registry_hive.get_key(registry_path)
        except RegistryKeyNotFoundException as ex:
            logger.debug('Did not find the key: {}'.format(ex))
            return
    else:
        name_key_entry = registry_hive.root

    if timeline and not output_path:
        click.secho('You must provide an output path if choosing timeline output!', fg='red')
        return

    if output_path:
        with open(output_path, 'w') as output_file:
            if timeline:
                csvwriter = csv.DictWriter(output_file, delimiter=',',
                                           quotechar='"', quoting=csv.QUOTE_MINIMAL,
                                           fieldnames=['timestamp', 'subkey_name', 'values_count', 'values'])
                csvwriter.writeheader()

            for subkey_count, entry in enumerate(get_filtered_subkeys(registry_hive, name_key_entry, fetch_values=not do_not_fetch_values, start_date=start_date, end_date=end_date)):
                if timeline:
                    csvwriter.writerow({
                        'subkey_name': entry.path,
                        'timestamp': entry.timestamp,
                        'values_count': entry.values_count,
                        'values': entry.values
                    })
                else:
                    output_file.write(json.dumps(attr.asdict(entry,), separators=(',', ':',), default=_normalize_subkey_fields))
                    output_file.write('\n')
    else:
        for subkey_count, entry in enumerate(registry_hive.recurse_subkeys(name_key_entry, as_json=True, fetch_values=not do_not_fetch_values)):
            click.secho(json.dumps(attr.asdict(entry), indent=4))

    click.secho(f"Completed in {time.monotonic() - start_time}s ({subkey_count} subkeys enumerated)")


@click.command()
@click.argument('hive_path', type=click.Path(exists=True, dir_okay=False, resolve_path=True), required=True)
@click.option('-o', 'output_path', type=click.Path(exists=False, dir_okay=False, resolve_path=True), required=True,
              help='Output path for plugins result')
@click.option('-p', '--plugins', type=click.STRING, required=False,
              help='A plugin or list of plugins to execute command separated')
@click.option('-t', '--hive-type', type=click.STRING, required=False,
              help='Specify a hive type, if it could not be identified for some reason')
@click.option('-r', '--partial_hive_path', type=click.STRING, required=False,
              help='The path from which the partial hive actually starts, for example: -t ntuser -r "/Software" '
                   'would mean this is actually a HKCU hive, starting from HKCU/Software')
@click.option('-v', '--verbose', is_flag=True, default=False, help='Verbosity')
def run_plugins(hive_path, output_path, plugins, hive_type, partial_hive_path, verbose):
    _setup_logging(verbose=verbose)
    registry_hive = RegistryHive(hive_path, hive_type=hive_type, partial_hive_path=partial_hive_path)
    click.secho('Loaded {} plugins'.format(len(PLUGINS)), fg='white')

    if plugins:
        plugin_names = {x.NAME for x in PLUGINS}
        plugins = plugins.split(',')
        plugins = set(plugins)
        if not plugins.issubset(plugin_names):
            click.secho('Invalid plugin names given: {}'.format(','.join(set(plugins) - plugin_names)), fg='red')
            click.secho('Use --help or -h to get list of plugins and their descriptions', fg='red')
            return

    # Run relevant plugins
    plugin_results = run_relevant_plugins(registry_hive, as_json=True, plugins=plugins)

    # If output path was set, dump results to disk
    if output_path:
        with open(output_path, 'w') as f:
            f.write(json.dumps(plugin_results, indent=4))
    else:
        print(json.dumps(plugin_results, indent=4))
    click.secho('Finished: {}/{} plugins matched the hive type'.format(len(plugin_results), len(PLUGINS)),
                fg='green')


@click.command()
def list_plugins():
    click.secho(tabulate([(x.NAME, x.COMPATIBLE_HIVE, x.DESCRIPTION) for x in PLUGINS],
                         headers=['Plugin Name', 'Compatible hive', 'Description']))


@click.command()
@click.argument('first_hive_path', type=click.Path(exists=True, dir_okay=False, resolve_path=True), required=True)
@click.argument('second_hive_path', type=click.Path(exists=True, dir_okay=False, resolve_path=True), required=True)
@click.option('-o', 'output_path', type=click.Path(exists=False, dir_okay=False, resolve_path=True), required=False)
@click.option('-v', '--verbose', is_flag=True, default=False, help='Verbosity')
def reg_diff(first_hive_path, second_hive_path, output_path, verbose):
    _setup_logging(verbose=verbose)
    REGDIFF_HEADERS = ['difference', 'first_hive', 'second_hive', 'description']

    found_differences = compare_hives(first_hive_path, second_hive_path, verbose=verbose)
    click.secho('Comparing {} vs {}'.format(os.path.basename(first_hive_path), os.path.basename(second_hive_path)))

    if output_path:
        with open(output_path, 'w') as csvfile:
            csvwriter = csv.writer(csvfile, delimiter='|', quoting=csv.QUOTE_MINIMAL)
            csvwriter.writerow(REGDIFF_HEADERS)
            for difference in found_differences:
                csvwriter.writerow(difference)
    else:
        click.secho(tabulate(found_differences, headers=REGDIFF_HEADERS,
                             tablefmt='fancy_grid'))
    click.secho(f'Detected {len(found_differences)} differences', fg='green')


@click.command()
@click.argument('hive_path', type=click.Path(exists=True, dir_okay=False, resolve_path=True), required=True)
@click.option('-p', 'primary_log_path', type=click.Path(exists=True, dir_okay=False, resolve_path=True), required=True)
@click.option('-s', 'secondary_log_path', type=click.Path(exists=True, dir_okay=False, resolve_path=True),
              required=False)
@click.option('-o', 'output_path', type=click.Path(exists=False, dir_okay=False, resolve_path=True), required=False)
@click.option('-v', '--verbose', is_flag=True, default=True, help='Verbosity')
def parse_transaction_log(hive_path, primary_log_path, secondary_log_path, output_path, verbose):
    _setup_logging(verbose=verbose)
    logger.info(f'Processing hive {hive_path} with transaction log {primary_log_path}')
    if secondary_log_path:
        logger.info(f'Processing hive {hive_path} with secondary transaction log {secondary_log_path}')

    restored_hive_path, recovered_dirty_pages_count = apply_transaction_logs(hive_path, primary_log_path,
                                                                             secondary_log_path=secondary_log_path,
                                                                             restored_hive_path=output_path,
                                                                             verbose=verbose)
    if recovered_dirty_pages_count:
        click.secho(
            f'Recovered {recovered_dirty_pages_count} dirty pages. Restored hive is at {restored_hive_path}',
            fg='green')



import logging

from inflection import underscore

from regipy import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

WINSCP_SAVED_SESSIONS_PATH = r'\Software\Martin Prikryl\WinSCP 2\Sessions'

class WinSCPSavedSessionsPlugin(Plugin):
    NAME = 'winscp_saved_sessions'
    DESCRIPTION = 'Retrieve list of WinSCP saved sessions'
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def _get_winscp_saved_sessions(self, subkey_path):
        try:
            sessions_sk = self.registry_hive.get_key(subkey_path)
        except RegistryKeyNotFoundException as ex:
            logger.error(ex)
            return

        for winscp_saved_session in sessions_sk.iter_subkeys():
            values = {x.name: x.value for x in
                      winscp_saved_session.iter_values(as_json=self.as_json)} if winscp_saved_session.values_count else {}
            self.entries.append({
                'timestamp': convert_wintime(winscp_saved_session.header.last_modified, as_json=self.as_json),
                'hive_name': 'HKEY_CURRENT_USER',
                'key_path': fr'HKEY_CURRENT_USER{subkey_path}\{winscp_saved_session.name}',
                **values
            })

    def run(self):
        self._get_winscp_saved_sessions(WINSCP_SAVED_SESSIONS_PATH)




import logging

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

TRACING_PATH = r'\Microsoft\Tracing'
X86_TRACING_PATH = r'\Wow6432Node' + TRACING_PATH


class RASTracingPlugin(Plugin):
    NAME = 'ras_tracing'
    DESCRIPTION = 'Retrieve list of executables using ras'
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def _get_installed_software(self, subkey_path):
        try:
            ras_subkey = self.registry_hive.get_key(subkey_path)
        except RegistryKeyNotFoundException as ex:
            logger.error(ex)
            return

        for entry in ras_subkey.iter_subkeys():
            timestamp = convert_wintime(entry.header.last_modified, as_json=self.as_json)
            self.entries.append({
                'key': subkey_path,
                'name': entry.name,
                'timestamp': timestamp
            })

    def run(self):
        self._get_installed_software(TRACING_PATH)
        self._get_installed_software(X86_TRACING_PATH)



"""
Windows machine domain name and SID extractor plugin
"""

import logging

from typing import Optional

from regipy.hive_types import SECURITY_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.structs import SID
from regipy.utils import convert_wintime
from regipy.security_utils import convert_sid

logger = logging.getLogger(__name__)

DOMAIN_NAME_PATH = r"\Policy\PolPrDmN"
DOMAIN_SID_PATH = r"\Policy\PolMachineAccountS"


class DomainSidPlugin(Plugin):
    """
    Windows machine domain name and SID extractor
    """

    NAME = "domain_sid"
    DESCRIPTION = "Get the machine domain name and SID"
    COMPATIBLE_HIVE = SECURITY_HIVE_TYPE

    def run(self) -> None:
        logger.debug("Started Machine Domain SID Plugin...")

        name_key = self.registry_hive.get_key(DOMAIN_NAME_PATH)

        # Primary Domain Name or Workgroup Name (binary-encoded and length-prefixed)
        name_value = name_key.get_value()

        # Skip UNICODE_STRING struct header and strip trailing \x0000
        domain_name = name_value[8:].decode('utf-16-le', errors='replace').rstrip('\x00')

        sid_key = self.registry_hive.get_key(DOMAIN_SID_PATH)

        # Domain SID value (binary-encoded)
        sid_value = sid_key.get_value()

        domain_sid: Optional[str] = None
        machine_sid: Optional[str] = None

        # The default key value is 0x00000000 (REG_DWORD) when
        # the Windows machine is not in an AD domain.
        # Otherwise, it contains the domain machine SID data
        # in the standard binary format (REG_BINARY).
        if isinstance(sid_value, bytes):
            parsed_sid = SID.parse(sid_value)
            domain_sid = convert_sid(parsed_sid, strip_rid=True)
            machine_sid = convert_sid(parsed_sid)

        self.entries.append(
            {
                "domain_name": domain_name,
                "domain_sid": domain_sid,
                "machine_sid": machine_sid,
                "timestamp": convert_wintime(
                    sid_key.header.last_modified, as_json=self.as_json
                ),
            }
        )



import logging

from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.plugins.system.external.ShimCacheParser import get_shimcache_entries

logger = logging.getLogger(__name__)

COMPUTER_NAME_PATH = r'Control\Session Manager'


class ShimCachePlugin(Plugin):
    NAME = 'shimcache'
    DESCRIPTION = 'Parse Shimcache artifact'
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug('Started Shim Cache Plugin...')

        for subkey_path in self.registry_hive.get_control_sets(COMPUTER_NAME_PATH):
            appcompat_cache = self.registry_hive.get_key(subkey_path).get_subkey('AppCompatCache')

            if not appcompat_cache:
                logger.info('No shimcache data was found')
                return

            shimcache = appcompat_cache.get_value('AppCompatCache')
            if shimcache:
                for entry in get_shimcache_entries(shimcache, as_json=self.as_json):
                    self.entries.append(entry)



import logging

from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import get_subkey_values_from_list, convert_wintime

logger = logging.getLogger(__name__)

IMAGE_FILE_EXECUTION_OPTIONS = r'\Microsoft\Windows NT\CurrentVersion\Image File Execution Options'


class ImageFileExecutionOptions(Plugin):
    NAME = 'image_file_execution_options'
    DESCRIPTION = 'Retrieve image file execution options - a persistence method'
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        image_file_execution_options = self.registry_hive.get_key(IMAGE_FILE_EXECUTION_OPTIONS)
        if image_file_execution_options.subkey_count:
            for subkey in image_file_execution_options.iter_subkeys():
                values = {x.name: x.value for x in
                          subkey.iter_values(as_json=self.as_json)} if subkey.values_count else {}
                self.entries.append({
                    'name': subkey.name,
                    'timestamp': convert_wintime(subkey.header.last_modified, as_json=self.as_json),
                    **values

                })



import logging

from regipy import RegistryKeyNotFoundException
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

X64_INSTALLED_SOFTWARE_PATH = r'\Microsoft\Windows\CurrentVersion\Uninstall'
X86_INSTALLED_SOFTWARE_PATH = r'\Wow6432Node' + X64_INSTALLED_SOFTWARE_PATH


class InstalledProgramsSoftwarePlugin(Plugin):
    NAME = 'installed_programs_software'
    DESCRIPTION = 'Retrieve list of installed programs and their install date from the SOFTWARE Hive'
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def _get_installed_software(self, subkey_path):
        try:
            uninstall_sk = self.registry_hive.get_key(subkey_path)
        except RegistryKeyNotFoundException as ex:
            logger.error(ex)
            return

        for installed_program in uninstall_sk.iter_subkeys():
            values = {x.name: x.value for x in
                      installed_program.iter_values(as_json=self.as_json)} if installed_program.values_count else {}
            self.entries.append({
                'service_name': installed_program.name,
                'timestamp': convert_wintime(installed_program.header.last_modified, as_json=self.as_json),
                'registry_path': subkey_path,
                **values

            })

    def run(self):
        self._get_installed_software(X64_INSTALLED_SOFTWARE_PATH)
        self._get_installed_software(X86_INSTALLED_SOFTWARE_PATH)




import logging

from regipy import RegistryKeyNotFoundException, convert_wintime, NoRegistrySubkeysException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

TSCLIENT_HISTORY_PATH = r'\Software\Microsoft\Terminal Server Client\Servers'


class TSClientPlugin(Plugin):
    NAME = 'terminal_services_history'
    DESCRIPTION = 'Retrieve history of RDP connections'
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        try:
            tsclient_subkey = self.registry_hive.get_key(TSCLIENT_HISTORY_PATH)
        except (RegistryKeyNotFoundException, NoRegistrySubkeysException) as ex:
            logger.error(ex)
            return

        for server in tsclient_subkey.iter_subkeys():
            self.entries.append({
                'server': server.name,
                'last_connection': convert_wintime(server.header.last_modified, as_json=self.as_json),
                'username_hint': server.get_value('UsernameHint')
            })




import logging

from regipy.exceptions import NoRegistrySubkeysException, RegistryKeyNotFoundException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

USBSTOR_KEY_PATH = r'Enum\USBSTOR'
DISK_GUID_PATH = r'Device Parameters\Partmgr'
PROPERTIES_NAME_GUID = r'{540b947e-8b40-45bc-a8a2-6a0b894cbda2}'
PROPERTIES_DATES_GUID = r'{83da6326-97a6-4088-9453-a1923f573b29}'
DEVICE_NAME_KEY = '0004'
FIRST_INSTALLED_TIME_KEY = '0065'
LAST_CONNECTED_TIME_KEY = '0066'
LAST_REMOVED_TIME_KEY = '0067'
LAST_INSTALLED_TIME_KEY = '0064'


class USBSTORPlugin(Plugin):
    NAME = 'usbstor_plugin'
    DESCRIPTION = "Parse the connected USB devices history"
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        try:
            for subkey_path in self.registry_hive.get_control_sets(USBSTOR_KEY_PATH):
                usbstor_key = self.registry_hive.get_key(subkey_path)
                for usbstor_drive in usbstor_key.iter_subkeys():
                    try:
                        disk, manufacturer, title, version = usbstor_drive.name.split('&')
                    except ValueError:
                        manufacturer, title, version = None, None, None

                    for serial_subkey in usbstor_drive.iter_subkeys():
                        timestamp = convert_wintime(serial_subkey.header.last_modified, as_json=self.as_json)
                        serial_number = serial_subkey.name
                        key_path = rf'{subkey_path}\{usbstor_drive.name}\{serial_number}'

                        try:
                            device_guid_key = self.registry_hive.get_key(
                                rf'{subkey_path}\{usbstor_drive.name}\{serial_number}\{DISK_GUID_PATH}'
                            )
                            disk_guid = device_guid_key.get_value('DiskId')
                        except RegistryKeyNotFoundException:
                            disk_guid = None

                        properties_subkey = serial_subkey.get_subkey('Properties')
                        if not properties_subkey:
                            device_name, first_installed_time, last_connected_time, last_removed_time, \
                                last_installed_time = None, None, None, None, None
                        else:
                            try:
                                device_name_guid_key = properties_subkey.get_subkey(PROPERTIES_NAME_GUID)
                                device_name_key = device_name_guid_key.get_subkey(DEVICE_NAME_KEY)
                                device_name = device_name_key.get_value()
                            except (RegistryKeyNotFoundException, NoRegistrySubkeysException):
                                device_name = None

                            first_installed_time, last_connected_time, last_removed_time, \
                                last_installed_time = None, None, None, None

                            dates_subkey = properties_subkey.get_subkey(PROPERTIES_DATES_GUID, raise_on_missing=False)
                            if dates_subkey:
                                first_installed_key = dates_subkey.get_subkey(FIRST_INSTALLED_TIME_KEY, raise_on_missing=False)
                                if first_installed_key:
                                    first_installed_time = first_installed_key.get_value(as_json=self.as_json)

                                last_connected_key = dates_subkey.get_subkey(LAST_CONNECTED_TIME_KEY,
                                                                             raise_on_missing=False)
                                if last_connected_key:
                                    last_connected_time = last_connected_key.get_value(as_json=self.as_json)

                                last_removed_key = dates_subkey.get_subkey(LAST_REMOVED_TIME_KEY,
                                                                           raise_on_missing=False)
                                if last_removed_key:
                                    last_removed_time = last_removed_key.get_value(as_json=self.as_json)

                                last_installed_key = dates_subkey.get_subkey(LAST_INSTALLED_TIME_KEY,
                                                                             raise_on_missing=False)
                                if last_installed_key:
                                    last_installed_time = last_installed_key.get_value(as_json=self.as_json)

                        self.entries.append({
                            'last_write': timestamp,
                            'key_path': key_path,
                            'last_connected': last_connected_time,
                            'last_removed': last_removed_time,
                            'first_installed': first_installed_time,
                            'last_installed': last_installed_time,
                            'serial_number': serial_number,
                            'device_name': device_name,
                            'disk_guid': disk_guid,
                            'manufacturer': manufacturer,
                            'version': version,
                            'title': title,
                        })

        except RegistryKeyNotFoundException as ex:
            logger.error(f'Could not find {self.NAME} plugin data at: {USBSTOR_KEY_PATH}: {ex}')



import logging

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

NETWORK_DRIVES = r'\Network'


class NetworkDrivesPlugin(Plugin):
    NAME = 'network_drives_plugin'
    DESCRIPTION = "Parse the user's mapped network drives"
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        try:
            network_drives = self.registry_hive.get_key(NETWORK_DRIVES)
            for mapped_drive in network_drives.iter_subkeys():
                timestamp = convert_wintime(mapped_drive.header.last_modified, as_json=self.as_json)
                self.entries.append({
                    'last_write': timestamp,
                    'drive_letter': mapped_drive.name,
                    'network_path': mapped_drive.get_value('RemotePath')
                })

        except RegistryKeyNotFoundException as ex:
            logger.error(f'Could not find {self.NAME} plugin data at: {NETWORK_DRIVES}: {ex}')


"""
Windows Boot Configuration Data (BCD) boot entry list plugin
"""

import logging
import uuid

from typing import Union

from regipy.registry import NKRecord
from regipy.hive_types import BCD_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

# BCD object store key path
# See https://www.geoffchappell.com/notes/windows/boot/bcd/objects.htm
BCD_OBJECTS_PATH = r"\Objects"

# Relevant BCD object element types:

# BcdLibraryDevice_ApplicationDevice
ELEM_TYPE_APPLICATION_DEVICE = 0x11000001
# BcdLibraryString_ApplicationPath
ELEM_TYPE_APPLICATION_PATH = 0x12000002
# BcdLibraryString_Description
ELEM_TYPE_DESCRIPTION = 0x12000004


def _get_element_by_type(obj_key: NKRecord, datatype: int) -> Union[str, bytes, None]:
    """
    Retrieves stored BCD object elements by their datatype.

    See https://www.geoffchappell.com/notes/windows/boot/bcd/elements.htm
    """

    # The BCD object attributes are stored as "elements" instead of normal values
    elements_key = obj_key.get_subkey("Elements", raise_on_missing=False)
    if elements_key.subkey_count == 0:
        return None

    elem_key = elements_key.get_subkey("%08X" % datatype, raise_on_missing=False)
    if elem_key is None:
        return None

    return elem_key.get_value("Element")


class BootEntryListPlugin(Plugin):
    """
    Windows Boot Configuration Data (BCD) boot entry list extractor
    """

    NAME = "boot_entry_list"
    DESCRIPTION = "List the Windows BCD boot entries"
    COMPATIBLE_HIVE = BCD_HIVE_TYPE

    def run(self) -> None:
        logger.debug("Started Boot Entry List Plugin...")

        objects_key = self.registry_hive.get_key(BCD_OBJECTS_PATH)

        for obj_key in objects_key.iter_subkeys():
            desc_key = obj_key.get_subkey("Description")
            # Object type defines the boot entry features
            desc_type = desc_key.get_value("Type")

            # The remaining boot entry attributes are stored as object elements
            desc_name = _get_element_by_type(obj_key, ELEM_TYPE_DESCRIPTION)
            path_name = _get_element_by_type(obj_key, ELEM_TYPE_APPLICATION_PATH)
            device_data = _get_element_by_type(obj_key, ELEM_TYPE_APPLICATION_DEVICE)

            # Filter out objects that do not look like boot entries
            if desc_name is None or path_name is None or device_data is None:
                continue

            # TODO: Figure out the device data blob format
            if not isinstance(device_data, bytes) or len(device_data) < 72:
                continue

            # TODO: Figure out how non-GPT partitions are encoded
            gpt_part_guid = str(uuid.UUID(bytes_le=device_data[32:48]))
            gpt_disk_guid = str(uuid.UUID(bytes_le=device_data[56:72]))

            entry_type = "0x%08X" % desc_type if self.as_json else desc_type

            self.entries.append(
                {
                    "guid": obj_key.name,
                    "type": entry_type,
                    "name": desc_name,
                    "gpt_disk": gpt_disk_guid,
                    "gpt_partition": gpt_part_guid,
                    "image_path": path_name,
                    "timestamp": convert_wintime(
                        obj_key.header.last_modified, as_json=self.as_json
                    ),
                }
            )



import logging
from inflection import underscore

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import NTUSER_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)


TYPED_PATHS_KEY_PATH = r'\Software\Microsoft\Windows\CurrentVersion\Explorer\TypedPaths'


class TypedPathsPlugin(Plugin):
    NAME = 'typed_paths'
    DESCRIPTION = 'Retrieve the typed Paths from the history'
    COMPATIBLE_HIVE = NTUSER_HIVE_TYPE

    def run(self):
        try:
            subkey = self.registry_hive.get_key(TYPED_PATHS_KEY_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(f'Could not find {self.NAME} plugin data at: {TYPED_PATHS_KEY_PATH}: {ex}')
            return None

        self.entries = {
            'last_write': convert_wintime(subkey.header.last_modified, as_json=self.as_json),
            'entries': [{underscore(x.name): x.value} for x in subkey.iter_values(as_json=self.as_json)]
        }



import logging

from construct import Int64ul

from regipy.exceptions import RegistryKeyNotFoundException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

BAM_PATH = [r'Services\bam\UserSettings',
            r'Services\bam\state\UserSettings']


class BAMPlugin(Plugin):
    NAME = 'background_activity_moderator'
    DESCRIPTION = 'Get the computer name'
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug('Started Computer Name Plugin...')

        for path in BAM_PATH:
            try:
                for subkey_path in self.registry_hive.get_control_sets(path):
                    subkey = self.registry_hive.get_key(subkey_path)
                    for sid_subkey in subkey.iter_subkeys():

                        sid = sid_subkey.name
                        logger.debug(f'Parsing BAM for {sid}')
                        sequence_number = None
                        version = None
                        entries = []

                        for value in sid_subkey.get_values(trim_values=self.trim_values):
                            if value.name == 'SequenceNumber':
                                sequence_number = value.value
                            elif value.name == 'Version':
                                version = value.value
                            else:
                                entries.append({
                                    'executable': value.name,
                                    'timestamp': convert_wintime(Int64ul.parse(value.value), as_json=self.as_json)
                                })

                        self.entries.extend([
                            {
                                'sequence_number': sequence_number,
                                'version': version,
                                'sid': sid,
                                'key_path': f'{subkey_path}\\{sid}',
                                **x
                            } for x in entries]
                        )
            except RegistryKeyNotFoundException as ex:
                logger.error(ex)



import logging
import attr

from regipy.exceptions import RegistryValueNotFoundException
from regipy.hive_types import SYSTEM_HIVE_TYPE
from regipy.plugins.plugin import Plugin
from regipy.utils import convert_wintime

logger = logging.getLogger(__name__)

COMPUTER_NAME_PATH = r'Control\ComputerName\ComputerName'


class ComputerNamePlugin(Plugin):
    NAME = 'computer_name'
    DESCRIPTION = 'Get the computer name'
    COMPATIBLE_HIVE = SYSTEM_HIVE_TYPE

    def run(self):
        logger.debug('Started Computer Name Plugin...')

        for subkey_path in self.registry_hive.get_control_sets(COMPUTER_NAME_PATH):
            subkey = self.registry_hive.get_key(subkey_path)

            try:
                self.entries.append({
                    'name': subkey.get_value('ComputerName', as_json=self.as_json),
                    'timestamp': convert_wintime(subkey.header.last_modified, as_json=self.as_json)
                })
            except RegistryValueNotFoundException as ex:
                continue



# Andrew Davis, andrew.davis@mandiant.com
# Copyright 2012 Mandiant
#
# Mandiant licenses this file to you under the Apache License, Version
# 2.0 (the "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at:
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.
#
# Identifies and parses Application Compatibility Shim Cache entries for forensic data.


# Original Mandiant shim cache parser, ported to Python 3
# Identifies and parses Application Compatibility Shim Cache entries for forensic data.

import datetime
import io as sio
import struct

import logging
import pytz

CACHE_MAGIC_NT5_2 = 0xbadc0ffe
CACHE_HEADER_SIZE_NT5_2 = 0x8
NT5_2_ENTRY_SIZE32 = 0x18
NT5_2_ENTRY_SIZE64 = 0x20

# Values used by Windows 6.1 (Win7 and Server 2008 R2)
CACHE_MAGIC_NT6_1 = 0xbadc0fee
CACHE_HEADER_SIZE_NT6_1 = 0x80
NT6_1_ENTRY_SIZE32 = 0x20
NT6_1_ENTRY_SIZE64 = 0x30
CSRSS_FLAG = 0x2

# Values used by Windows 5.1 (WinXP 32-bit)
WINXP_MAGIC32 = 0xdeadbeef
WINXP_HEADER_SIZE32 = 0x190
WINXP_ENTRY_SIZE32 = 0x228
MAX_PATH = 520

# Values used by Windows 8
WIN8_STATS_SIZE = 0x80
WIN8_MAGIC = b'00ts'

# Magic value used by Windows 8.1
WIN81_MAGIC = b'10ts'

# Values used by Windows 10
WIN10_STATS_SIZE = 0x30
WIN10_MAGIC = b'10ts'
CACHE_HEADER_SIZE_NT6_4 = 0x30
CACHE_MAGIC_NT6_4 = 0x30

BAD_ENTRY_DATA = 'N/A'
G_VERBOSE = False
G_USE_BOM = False
OUTPUT_HEADER = ["Last Modified", "Last Update", "Path", "File Size", "Exec Flag"]

logger = logging.getLogger(__name__)

# Shim Cache format used by Windows 5.2 and 6.0 (Server 2003 through Vista/Server 2008)
class CacheEntryNt5(object):
    def __init__(self, is_32_bit, data=None):
        self.w_length = None
        self.w_maximum_length = None
        self.offset = None
        self.dw_low_date_time = None
        self.dw_high_date_time = None
        self.dw_file_size_low = None
        self.dw_file_size_high = None
        self.is_32_bit = is_32_bit
        if data is not None:
            self.update(data)

    def update(self, data):
        if self.is_32_bit:
            entry = struct.unpack('<2H 3L 2L', data)
        else:
            entry = struct.unpack('<2H 4x Q 2L 2L', data)
        self.w_length = entry[0]
        self.w_maximum_length = entry[1]
        self.offset = entry[2]
        self.dw_low_date_time = entry[3]
        self.dw_high_date_time = entry[4]
        self.dw_file_size_low = entry[5]
        self.dw_file_size_high = entry[6]

    def size(self):

        if self.is_32_bit:
            return NT5_2_ENTRY_SIZE32
        else:
            return NT5_2_ENTRY_SIZE64


# Shim Cache format used by Windows 6.1 (Win7 through Server 2008 R2).
class CacheEntryNt6(object):
    def __init__(self, is_32_bit, data=None):
        self.w_length = None
        self.w_maximum_length = None
        self.offset = None
        self.dw_low_date_time = None
        self.dw_high_date_time = None
        self.file_flags = None
        self.flags = None
        self.blob_size = None
        self.blob_offset = None
        self.is_32_bit = is_32_bit
        if data is not None:
            self.update(data)

    def update(self, data):
        if self.is_32_bit:
            entry = struct.unpack('<2H 7L', data)
        else:
            entry = struct.unpack('<2H 4x Q 4L 2Q', data)
        self.w_length = entry[0]
        self.w_maximum_length = entry[1]
        self.offset = entry[2]
        self.dw_low_date_time = entry[3]
        self.dw_high_date_time = entry[4]
        self.file_flags = entry[5]
        self.flags = entry[6]
        self.blob_size = entry[7]
        self.blob_offset = entry[8]

    def size(self):
        if self.is_32_bit:
            return NT6_1_ENTRY_SIZE32
        else:
            return NT6_1_ENTRY_SIZE64


# Convert FILETIME to datetime.
# Based on http://code.activestate.com/recipes/511425-filetime-to-datetime/
def convert_filetime(dw_low_date_time, dw_high_date_time):
    try:
        date = datetime.datetime(1601, 1, 1, 0, 0, 0)
        temp_time = dw_high_date_time
        temp_time <<= 32
        temp_time |= dw_low_date_time
        return pytz.utc.localize(date + datetime.timedelta(microseconds=temp_time / 10))
    except OverflowError:
        return None


# Return a unique list while preserving ordering.
def unique_list(li):
    ret_list = []
    for entry in li:
        if entry not in ret_list:
            ret_list.append(entry)
    return ret_list


# Read the Shim Cache format, return a list of last modified dates/paths.
def get_shimcache_entries(cachebin, as_json=False):

    if len(cachebin) < 16:
        # Data size less than minimum header size.
        return None

    # Get the format type
    magic = struct.unpack("<L", cachebin[0:4])[0]

    # This is a Windows 2k3/Vista/2k8 Shim Cache format,
    if magic == CACHE_MAGIC_NT5_2:

        # Shim Cache types can come in 32-bit or 64-bit formats. We can
        # determine this because 64-bit entries are serialized with u_int64
        # pointers. This means that in a 64-bit entry, valid UNICODE_STRING
        # sizes are followed by a NULL DWORD. Check for this here.
        test_size = struct.unpack("<H", cachebin[8:10])[0]
        test_max_size = struct.unpack("<H", cachebin[10:12])[0]
        if (test_max_size - test_size == 2 and
                struct.unpack("<L", cachebin[12:16])[0]) == 0:
            logger.debug("[+] Found 64bit Windows 2k3/Vista/2k8 Shim Cache data...")
            entry = CacheEntryNt5(False)
            yield from read_nt5_entries(cachebin, entry, as_json=as_json)

        # Otherwise it's 32-bit data.
        else:
            logger.debug("[+] Found 32bit Windows 2k3/Vista/2k8 Shim Cache data...")
            entry = CacheEntryNt5(True)
            yield from read_nt5_entries(cachebin, entry, as_json=as_json)

    # This is a Windows 7/2k8-R2 Shim Cache.
    elif magic == CACHE_MAGIC_NT6_1:
        test_size = (struct.unpack("<H", cachebin[CACHE_HEADER_SIZE_NT6_1:CACHE_HEADER_SIZE_NT6_1 + 2])[0])
        test_max_size = (struct.unpack("<H", cachebin[CACHE_HEADER_SIZE_NT6_1 + 2:CACHE_HEADER_SIZE_NT6_1 + 4])[0])

        # Shim Cache types can come in 32-bit or 64-bit formats.
        # We can determine this because 64-bit entries are serialized with
        # u_int64 pointers. This means that in a 64-bit entry, valid
        # UNICODE_STRING sizes are followed by a NULL DWORD. Check for this here.
        if (test_max_size - test_size == 2 and
                struct.unpack("<L", cachebin[CACHE_HEADER_SIZE_NT6_1 + 4:CACHE_HEADER_SIZE_NT6_1 + 8])[0]) == 0:
            logger.debug("[+] Found 64bit Windows 7/2k8-R2 Shim Cache data...")
            entry = CacheEntryNt6(False)
            yield from read_nt6_entries(cachebin, entry, as_json=as_json)
        else:
            logger.debug("[+] Found 32bit Windows 7/2k8-R2 Shim Cache data...")
            entry = CacheEntryNt6(True)
            yield from read_nt6_entries(cachebin, entry, as_json=as_json)

    # This is WinXP cache data
    elif magic == WINXP_MAGIC32:
        logger.debug("[+] Found 32bit Windows XP Shim Cache data...")
        yield from read_winxp_entries(cachebin, as_json=as_json)

    # Check the data set to see if it matches the Windows 8 format.
    elif len(cachebin) > WIN8_STATS_SIZE and cachebin[WIN8_STATS_SIZE:WIN8_STATS_SIZE + 4] == WIN8_MAGIC:
        logger.debug("[+] Found Windows 8/2k12 Apphelp Cache data...")
        yield from read_win8_entries(cachebin, WIN8_MAGIC, as_json=as_json)

    # Windows 8.1 will use a different magic dword, check for it
    elif len(cachebin) > WIN8_STATS_SIZE and cachebin[WIN8_STATS_SIZE:WIN8_STATS_SIZE + 4] == WIN81_MAGIC:
        logger.debug("[+] Found Windows 8.1 Apphelp Cache data...")
        yield from read_win8_entries(cachebin, WIN81_MAGIC, as_json=as_json)

    # Windows 10 will use a different magic dword, check for it
    elif len(cachebin) > WIN10_STATS_SIZE and cachebin[WIN10_STATS_SIZE:WIN10_STATS_SIZE + 4] == WIN10_MAGIC:
        logger.debug("[+] Found Windows 10 Apphelp Cache data...")
        yield from read_win10_entries(cachebin, WIN10_MAGIC, as_json=as_json)

    # Windows 10 creators update moved the damn magic 4 bytes forward...
    elif len(cachebin) > WIN10_STATS_SIZE and cachebin[WIN10_STATS_SIZE + 4:WIN10_STATS_SIZE + 8] == WIN10_MAGIC:
        logger.debug("[+] Found Windows 10 Apphelp Cache data... (creators update)")
        yield from read_win10_entries(cachebin, WIN10_MAGIC, creators_update=True, as_json=as_json)

    else:
        raise Exception('Got an unrecognized magic value of 0x{:x}... bailing'.format(magic))


# Read Windows 8/2k12/8.1 Apphelp Cache entry formats.
def read_win8_entries(bin_data, ver_magic, as_json=False):

    entry_meta_len = 12
    entry_list = []

    # Skip past the stats in the header
    cache_data = bin_data[WIN8_STATS_SIZE:]

    data = sio.BytesIO(cache_data)
    while data.tell() < len(cache_data):
        header = data.read(entry_meta_len)
        # Read in the entry metadata
        # Note: the crc32 hash is of the cache entry data
        magic, crc32_hash, entry_len = struct.unpack('<4sLL', header)

        # Check the magic tag
        if magic != ver_magic:
            raise Exception('Invalid version magic tag found: {}'.format(struct.unpack('I', magic)[0]))

        entry_data = sio.BytesIO(data.read(entry_len))

        # Read the path length
        path_len = struct.unpack('<H', entry_data.read(2))[0]
        if path_len == 0:
            path = 'None'
        else:
            path = entry_data.read(path_len).decode('utf-16le', 'replace')

        # Check for package data
        package_len = struct.unpack('<H', entry_data.read(2))[0]
        if package_len > 0:
            # Just skip past the package data if present (for now)
            entry_data.seek(package_len, 1)

        # Read the remaining entry data
        flags, unk_1, low_datetime, high_datetime, unk_2 = struct.unpack('<LLLLL', entry_data.read(20))

        # Check the flag set in CSRSS
        if flags & CSRSS_FLAG:
            exec_flag = 'True'
        else:
            exec_flag = 'False'

        last_mod_date = convert_filetime(low_datetime, high_datetime)

        yield {
            'last_mod_date': last_mod_date.isoformat() if as_json else last_mod_date,
            'path': path,
            'exec_flag': exec_flag
        }


# Read Windows 10 Apphelp Cache entry format
def read_win10_entries(bin_data, ver_magic, creators_update=False, as_json=False):

    entry_meta_len = 12
    entry_list = []

    # Skip past the stats in the header
    if creators_update:
        cache_data = bin_data[WIN10_STATS_SIZE + 4:]
    else:
        cache_data = bin_data[WIN10_STATS_SIZE:]

    data = sio.BytesIO(cache_data)
    while data.tell() < len(cache_data):
        header = data.read(entry_meta_len)
        # Read in the entry metadata
        # Note: the crc32 hash is of the cache entry data
        magic, crc32_hash, entry_len = struct.unpack('<4sLL', header)

        # Check the magic tag
        if magic != ver_magic:
            raise Exception('Invalid version magic tag found: {}'.format(struct.unpack('<I', magic)[0]))

        entry_data = sio.BytesIO(data.read(entry_len))

        # Read the path length
        path_len = struct.unpack('<H', entry_data.read(2))[0]
        if path_len == 0:
            path = 'None'
        else:
            path = entry_data.read(path_len).decode('utf-16le', 'replace')

        # Read the remaining entry data
        low_datetime, high_datetime = struct.unpack('<LL', entry_data.read(8))

        # Skip the unrecognized Microsoft App entry format for now
        if not (low_datetime+high_datetime):
            continue
        else:
            last_mod_date = convert_filetime(low_datetime, high_datetime)

        yield {
            'last_mod_date': last_mod_date.isoformat() if as_json else last_mod_date,
            'path': path
        }


# Read Windows 2k3/Vista/2k8 Shim Cache entry formats.
def read_nt5_entries(bin_data, entry, as_json=False):
    entry_list = []
    contains_file_size = False
    entry_size = entry.size()

    num_entries = struct.unpack('<L', bin_data[4:8])[0]
    if num_entries == 0:
        return None

    # On Windows Server 2008/Vista, the filesize is swapped out of this
    # structure with two 4-byte flags. Check to see if any of the values in
    # "dwFileSizeLow" are larger than 2-bits. This indicates the entry contained file sizes.
    for offset in range(CACHE_HEADER_SIZE_NT5_2, (num_entries * entry_size) + CACHE_HEADER_SIZE_NT5_2,
                        entry_size):

        entry.update(bin_data[offset:offset + entry_size])

        if entry.dw_file_size_low > 3:
            contains_file_size = True
            break

    # Now grab all the data in the value.
    for offset in range(CACHE_HEADER_SIZE_NT5_2, (num_entries * entry_size) + CACHE_HEADER_SIZE_NT5_2,
                        entry_size):

        entry.update(bin_data[offset:offset + entry_size])

        last_mod_date = convert_filetime(entry.dw_low_date_time, entry.dw_high_date_time)
        path = bin_data[entry.offset:entry.offset + entry.w_length].decode('utf-16le', 'replace')

        # It contains file size data.
        exec_flag = None
        if contains_file_size:
            yield {
                'last_mod_date': last_mod_date.isoformat() if as_json else last_mod_date,
                'path': path,
                'file_size': entry.dw_file_size_low
            }

        # It contains flags.
        else:
            # Check the flag set in CSRSS
            if entry.dw_file_size_low & CSRSS_FLAG:
                exec_flag = 'True'
            else:
                exec_flag = 'False'

            yield {
                'last_mod_date': last_mod_date.isoformat() if as_json else last_mod_date,
                'path': path,
                'exec_flag': exec_flag
            }


# Read the Shim Cache Windows 7/2k8-R2 entry format,
# return a list of last modifed dates/paths.
def read_nt6_entries(bin_data, entry, as_json=False):
    entry_size = entry.size()
    num_entries = struct.unpack('<L', bin_data[4:8])[0]

    if num_entries == 0:
        return None

    # Walk each entry in the data structure.
    for offset in range(CACHE_HEADER_SIZE_NT6_1, num_entries * entry_size + CACHE_HEADER_SIZE_NT6_1, entry_size):

        entry.update(bin_data[offset:offset + entry_size])
        last_mod_date = convert_filetime(entry.dw_low_date_time, entry.dw_high_date_time)
        path = bin_data[entry.offset:entry.offset + entry.w_length].decode('utf-16le', 'replace')

        # Test to see if the file may have been executed.
        if entry.file_flags & CSRSS_FLAG:
            exec_flag = 'True'
        else:
            exec_flag = 'False'

        yield {
            'last_mod_date': last_mod_date.isoformat() if as_json else last_mod_date,
            'path': path,
            'exec_flag': exec_flag
        }


# Read the WinXP Shim Cache data. Some entries can be missing data but still
# contain useful information, so try to get as much as we can.
def read_winxp_entries(bin_data, as_json=False):
    entry_list = []

    num_entries = struct.unpack('<L', bin_data[8:12])[0]
    if num_entries == 0:
        return None

    for offset in range(WINXP_HEADER_SIZE32,
                        (num_entries * WINXP_ENTRY_SIZE32) + WINXP_HEADER_SIZE32, WINXP_ENTRY_SIZE32):
        # No size values are included in these entries, so search for utf-16 terminator.
        path_len = bin_data[offset:offset + (MAX_PATH + 8)].find(b'\x00\x00')

        # if path is corrupt, procede to next entry.
        if path_len == 0:
            continue
        path = bin_data[offset:offset + path_len + 1].decode('utf-16le')

        if len(path) == 0:
            continue

        entry_data = (offset + (MAX_PATH + 8))

        # Get last mod time.
        last_mod_time = struct.unpack('<2L', bin_data[entry_data:entry_data + 8])
        last_mod_time = convert_filetime(last_mod_time[0],last_mod_time[1])

        # Get last file size.
        file_size = struct.unpack('<2L', bin_data[entry_data + 8:entry_data + 16])[0]
        if file_size == 0:
            file_size = BAD_ENTRY_DATA

        # Get last update time.
        exec_time = struct.unpack('<2L', bin_data[entry_data + 16:entry_data + 24])
        exec_time = convert_filetime(exec_time[0],exec_time[1])

        yield {
            'last_mod_time': last_mod_time.isoformat() if as_json else last_mod_time,
            'exec_time': exec_time.isoformat() if as_json else exec_time,
            'path': path,
            'file_size': file_size
        }


def parse_output(output):
    new_output_list = list()
    for row in output:
        exec_flag = False
        if row[4] == 'True':
            exec_flag = True
        entry = {'last_mod_date': row[0], 'last_update': row[1], 'path': row[2], 'file_size': row[3],
                 'exec_flag': exec_flag}
        new_output_list.append(entry)
    return new_output_list



import logging

from regipy import RegistryKeyNotFoundException, convert_wintime
from regipy.hive_types import SOFTWARE_HIVE_TYPE
from regipy.plugins.plugin import Plugin

logger = logging.getLogger(__name__)

CLASSES_INSTALLER_PATH = r'\Classes\Installer\Products'


class SoftwareClassesInstallerPlugin(Plugin):
    NAME = 'software_classes_installer'
    DESCRIPTION = 'List of installed software from SOFTWARE hive'
    COMPATIBLE_HIVE = SOFTWARE_HIVE_TYPE

    def run(self):
        try:
            installer_subkey = self.registry_hive.get_key(CLASSES_INSTALLER_PATH)
        except RegistryKeyNotFoundException as ex:
            logger.error(ex)
            return

        for entry in installer_subkey.iter_subkeys():
            identifier = entry.name
            timestamp = convert_wintime(entry.header.last_modified, as_json=self.as_json)
            product_name = entry.get_value('ProductName')
            self.entries.append({
                'identifier': identifier,
                'timestamp': timestamp,
                'product_name': product_name,
                'is_hidden': product_name is None
            })


